#!/usr/bin/env python
# coding: utf-8

from mavisim import AstromCalibSimAna, AstromCalibSimE2E
from astropy.io import ascii
import numpy as np
import pytest

def eval_astromsim(astrom_sim):
    # Values to evaluate distortion functions with. Can be any arangement of 
    # coordinates within the science field.
    nsamp = 31
    xx,yy = np.meshgrid(np.linspace(-15,15,nsamp),np.linspace(-15,15,nsamp))
    xx = xx.flatten()
    yy = yy.flatten()
    rr = (xx**2+yy**2)**0.5
    xx = xx[rr<=15.0]
    yy = yy[rr<=15.0]

    # interpolate input distortions at xx/yy coordinates
    input_dist_xx,input_dist_yy = astrom_sim.input_dist(xx,yy)

    # interpolate recovered distortions at xx/yy coordinates
    recovered_dist_xx,recovered_dist_yy = astrom_sim.recovered_dist(xx,yy)

    # interpolate residual distortions at xx/yy coordinates
    residual_dist_xx,residual_dist_yy = astrom_sim.residual_dist(xx,yy)
    # remove tt and plate scale since not in the requirements
    X = np.array([xx,yy,xx*0+1]).T
    coeffs = np.linalg.solve(X.T @ X,X.T) @ np.c_[residual_dist_xx,residual_dist_yy]
    filt = np.eye(xx.shape[0])- (X @ np.linalg.solve(X.T @ X,X.T))

    input_dist_xx = filt @ input_dist_xx
    input_dist_yy = filt @ input_dist_yy
    recovered_dist_xx = filt @ recovered_dist_xx
    recovered_dist_yy = filt @ recovered_dist_yy
    residual_dist_xx = filt @ residual_dist_xx
    residual_dist_yy = filt @ residual_dist_yy

    # Evaluate astrometric error between objects
    n_objects    = 4000
    objects      = np.random.random([n_objects,2])*30-15
    filt_objects = ((objects**2).sum(axis=1)**0.5)<=15.0
    objects      = objects[filt_objects,:]
    n_objects    = objects.shape[0]
    dists = []
    errs  = []
    true_pos = objects.copy()
    distort  = np.array([astrom_sim.residual_dist(ob[0],ob[1]) for ob in objects]) - objects @ coeffs[:2,:]
    distance = ((true_pos[:,None,:]-true_pos[None,:,:])**2).sum(axis=-1)**0.5
    errors = ((distort[:,None,:]-distort[None,:,:])**2).sum(axis=-1)**0.5
    indices = np.tril_indices(distance.shape[0],1)
    dists = distance[indices]
    errs = errors[indices]
    rel_err = errs[dists<1]
    rel_err_n = (dists<1).sum()
    print(f"  rel astrometric error: {rel_err.std()*1e6:7.3f} uas rms")
    print(f"           err < 150uas: {(rel_err < 150e-6).sum() * 100 / rel_err_n :6.2f}%")
    print(f"           err <  50uas: {(rel_err < 50e-6).sum() * 100 / rel_err_n :6.2f}%")
    abs_err = errs[:]
    abs_err_n = errs.shape[0]
    print(f"  abs astrometric error: {abs_err.std()*1e6:7.3f} uas rms")
    print(f"          err < 2000uas: {(abs_err < 2000e-6).sum() * 100 / abs_err_n :6.2f}%")
    print(f"          err <  400uas: {(abs_err < 400e-6).sum() * 100 / abs_err_n :6.2f}%")
    return errs

ana_testdata = [
    (
        0.01, # shift
        5, # npoly
        0.0, # noise
        20e-6 # rms performance threshold
    ),
    (
        0.3,
        5,
        10e-6,
        50e-6
    ),
    (
        0.3,
        5,
        20e-6,
        200e-6
    ),
]
@pytest.mark.parametrize("dx,n_poly,noise,thresh",ana_testdata)
def test_astromsim_analytical(dx,n_poly,noise,thresh):
    # input true distortion field:
    static_distort =  ascii.read("tests/test_StaticDistort")

    # create the astrometry simulator object:
    astrom_sim = AstromCalibSimAna(
        static_distort,           # astropy parsed input file with appropriate headers
        centroid_noise_std=noise, # centroiding noise in arcsec.
        dx=dx, dy=dx,             # Shift applied to mask for differential method.
        n_poly=n_poly)            # max order of polynomial to fit with.
    errs = eval_astromsim(astrom_sim)
    assert errs.std() < thresh

def noise_func(image,flux):
        image = np.random.poisson(image*flux)*1.0
        return image

e2e_testdata = [
    (
        0.3, # shift
        5, # npoly
        lambda x : x, # noise func
        20e-6 # rms performance threshold
    ),
    (
        0.3,
        5,
        lambda x : noise_func(x,6e5),
        50e-6
    ),
    (
        0.3,
        5,
        lambda x : noise_func(x,1e5),
        200e-6
    ),
]
@pytest.mark.parametrize("dx,n_poly,noise_func,thresh",e2e_testdata)
def test_astromsim_e2e(dx,n_poly,noise_func,thresh):
    # input true distortion field:
    static_distort =  ascii.read("tests/test_StaticDistort")
    # create the astrometry simulator object:
    astrom_sim = AstromCalibSimE2E(static_distort,
        num_pin_x=40, dx=dx, dy=dx, n_poly=n_poly,
        noise_fun=noise_func)
    errs = eval_astromsim(astrom_sim)
    assert errs.std() < thresh


if __name__ == "__main__":
    test_astromsim_analytical(*ana_testdata[0])
    test_astromsim_e2e(*e2e_testdata[0])


import pytest
from mavisim import ImageGenerator, Source
import numpy as np
import astropy.io.fits as fits
import tests.test_parameters as input_par
from importlib import reload

GAUSS_PSF = "tests/test_psf_gauss.fits"
E2E_PSF = "tests/test_psf_e2e.fits"


def make_source(n_sources=1):
    """Create a source table from the test_parameters file

    Returns:
        src : (mavisim.Source) source object
    """
    reload(input_par)
    for _ in range(n_sources - 1):
        input_par.input_cat.add_row(input_par.input_cat[0])
    exp_time = 10
    src = Source(input_par=input_par, exp_time=exp_time, static_dist=False)
    src.build_source()
    return src


def make_image(width=8000, padding=1000, rebin=2):
    """Create an image from the test source object and some default parameters

    Returns:
        im : (np.ndarray) image generated by mavisim
        src: (mavisim.Source) source object used to generate image
    """
    src = make_source()
    pixsize = 3.75e-3
    image_gen = ImageGenerator(array_width_pix=width + 2 * padding, source=src,
                               psfs_file=GAUSS_PSF, pixsize=pixsize,
                               gauss_width_pix=34, which_psf=0, norm_psf=True)
    image_gen.main()
    im = image_gen.get_rebinned_cropped(rebin_factor=rebin, cropped_width_as=width * pixsize)
    return im, src


def get_psf_cog(filename):
    """take a PSF from a fits file and return its centroid

    Returns:
        cog: (np.ndarray) (2,) array containing CoG in x and y, in pixel units
    """
    psf = fits.open(filename)[1].data
    cog = get_cog(psf)
    return cog


def get_cog(image):
    """returns the centre of gravity of an image in pixel units, here is where
    we define the coordinates of all subsequent operations. By convention, we
    propose that the (0,0) coordinate is in the absolute centre of the image.

    For example, a symmetric source placed at (0,0) pixels in an square image
    with an even number of pixels in each dimension (e.g., 4000x4000 pixels)
    will have uniform intensity across the 2x2 sub grid of pixels in the centre
    of the image. In numpy conventions, this is to say that:
        image.shape = (4000,4000)
        i = image[1999,1999]
        assert image[1999,2000] == i
        assert image[2000,1999] == i
        assert image[2000,2000] == i
        s = image[:2000,:2000].sum()
        assert image[2000:,:2000].sum() == s
        assert image[:2000,2000:].sum() == s
        assert image[2000:,2000:].sum() == s

    If the image were to have an odd number of pixels in each dimension (e.g., 3x3),
    then a symmetric object placed at (0,0) pixels should fall "in the centre" of the
    central pixel, and the 4 adjacent pixels (directly above,below,left,right of the
    centre) should have equal intensities to each other.

    This definition is consistent with the philosophy that pixels are "bins" of the
    continuous space, rather than discrete samples of the continuous space. So the
    full width of the image corresponds to exactly the number of pixels along the x
    dimension (N) multiplied by the pixel scale. Conversely, in the discrete sampling
    philosophy, there is a fencepost issue where the full width of the image is N-1
    multiplied by the pixel scale. This is a subtle difference but we're dealing with
    sub-pixel astrometry so it seems important to get this right.
    """
    yy, xx = np.mgrid[:image.shape[0], :image.shape[1]].astype(np.float64)
    yy -= image.shape[0] / 2  # centering the coordinates
    yy += 0.5  # offset to achieve desired convention
    xx -= image.shape[1] / 2  # centering the coordinates
    xx += 0.5  # offset to achieve desired convention
    im_sum = image.sum()
    cog_x = (image * xx).sum() / im_sum
    cog_y = (image * yy).sum() / im_sum
    return np.array([cog_x, cog_y])


def get_cog_windowed(im: np.ndarray, offset_x: int, offset_y: int, width: int):
    """Crop a large image to a smaller region and return the cog in the large image
    coordinates

    Arguments:
        im : (np.ndarray) full size square 2d array to be cog'd
        offset_x : (int) offset in x (axis=1) to start the cropping of the image
        offset_y : (int) same but in y (axis=0)
        width : (int) width of region to crop

    Returns:
        cog : (np.ndarray) (2,) array containing x and y CoG values
    """
    cog = get_cog(im[offset_y:offset_y + width, offset_x:offset_x + width])
    cog = cog + np.array([offset_x, offset_y]) + width / 2 - im.shape[0] / 2
    return cog


def test_make_image():
    """Test the image maker runs without errors and produces the expected shape image
    """
    im, _ = make_image(width=8000, padding=1000, rebin=2)
    assert im.shape[0] == 4000
    assert im.shape[1] == 4000

    im, _ = make_image(width=2000, padding=100, rebin=1)
    assert im.shape[0] == 2000
    assert im.shape[1] == 2000


def test_photometry():
    """Verify that flux is conserved from source table to final image
    """
    im, src = make_image()
    assert np.abs((im.sum() - src.flux.sum()) / src.flux.sum()) < 0.001  # make sure less than 0.1% flux error per star


def test_cog_convention():
    """Verify that cog of an even and odd shape image produces the expected position
    """
    # even
    # central source
    im = np.array([
        [0, 0, 0, 0],
        [0, 7, 7, 0],
        [0, 7, 7, 0],
        [0, 0, 0, 0]
    ]).astype(np.float64)
    assert np.allclose(get_cog(im), np.array([0, 0]))
    # pos x-shifted source
    im = np.array([
        [0, 0, 0, 0],
        [0, 0, 2, 2],
        [0, 0, 2, 2],
        [0, 0, 0, 0]
    ]).astype(np.float64)
    assert np.allclose(get_cog(im), np.array([1, 0]))

    # odd
    # central source
    im = np.array([
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0]
    ]).astype(np.float64)
    assert np.allclose(get_cog(im), np.array([0, 0]))
    # neg y-shifted source
    im = np.array([
        [0, 0, 9, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0]
    ]).astype(np.float64)
    assert np.allclose(get_cog(im), np.array([0, -2]))


def test_init_pos_is_zero():
    """Verify that the default source object has a position of zero, since
    that is what should be specified in the source table.
    """
    im, _ = make_image()
    cog = get_cog(im)
    assert np.allclose(cog, np.array([0, 0]))


def get_gr_distributed_sources(npoints):
    """Create a set of source coordinates on the unit disc to be used for pseudo-randomly
    distributing test sources. The furthest point from the centre lies just inside the unit
    disc.

    Uses a set of points distributed by the golden ratio to try and avoid repetative pixel
    sampling and pass tests with a hidden bug.

    Returns:
        pos : (np.ndarray) (npoints,2) array of coordinates [[x1,y1],[x2,y2],...,[xnpoints,ynpoints]]
    """
    gr_theta = (3 - 5**0.5) * np.pi
    radius = (np.arange(npoints) / npoints)**0.5
    theta = (np.arange(npoints) * gr_theta)
    pos = np.array([radius * np.cos(theta), radius * np.sin(theta)]).T
    return pos


def test_psf_cog():
    """PSFs should be saved so that the average tip/tilt in the phase space has
    an expectation of zero, which corresponds to a shift of half a pixel with respect
    to the coordinate definition used elsewhere. This appears to be the default
    convention used in most AO simulation software.
    """
    psf_cog = get_psf_cog(GAUSS_PSF)
    assert np.allclose(psf_cog, np.array([0.5, 0.5]))


@pytest.mark.parametrize("psf_file,tolerance,rebin", [
    (GAUSS_PSF, 1e-7, 1),  # gaussian PSFs can tolerate a much higher precision
    (GAUSS_PSF, 1e-7, 2),
    (GAUSS_PSF, 1e-7, 4),
    (E2E_PSF, 1e-2, 1),  # e2e psfs have non-zero strength all the way to the edges
    (E2E_PSF, 1e-2, 2),  # so classical CoG algorithms are not perfect.
    (E2E_PSF, 1e-2, 4),
])
def test_astrometry(psf_file, tolerance, rebin):
    """Verify that an image with a single source can have its position
    recovered to a desired accuracy.

    This function works by generating many images with floating point positioned
    sources and performing a CoG over the entire image. If the image is oversampled,
    not truncated, and noiseless, this should be exactly the right value (from uuu
    Fourier transform theory). Gaussian PSFs behave this way, but e2e PSFs aren't
    perfectly zero for any window size, probably due to the finite precision FFT
    computations performed to make them. The Gaussian PSF is made directly in the
    image space so is truly zero at some distance from the centre.
    """
    psf_cog = get_psf_cog(psf_file) - np.array([0.5, 0.5])

    # make image and check cog
    src = make_source()
    err_xl = []
    err_yl = []
    max_rad = 5.0  # arcsec
    pixsize = 3.75e-3  # arcsec
    npoints = 10
    pos = max_rad * get_gr_distributed_sources(npoints)
    for new_pos in pos:
        src.gauss_pos[:] = new_pos.copy()
        image_gen = ImageGenerator(array_width_pix=5000, source=src, psfs_file=psf_file,
                                   pixsize=pixsize, gauss_width_pix=32, which_psf=0, norm_psf=True)
        image_gen.main()
        im = image_gen.get_rebinned_cropped(rebin_factor=rebin, cropped_width_as=15)
        cog_x, cog_y = get_cog(im)
        # measure error in pixels:
        err_xl.append(cog_x * rebin - new_pos[0] / pixsize - psf_cog[0])
        err_yl.append(cog_y * rebin - new_pos[1] / pixsize - psf_cog[1])
    err_x = np.array(err_xl)
    err_y = np.array(err_yl)
    assert err_x.std() < tolerance     # x error std is less than 1e-6 pixel (in fine scale)
    assert err_y.std() < tolerance     # y error std is less than 1e-6 pixel (in fine scale)
    assert np.abs(err_x.mean()) < tolerance  # x bias is less than 1e-6 pixel (in fine scale)
    assert np.abs(err_y.mean()) < tolerance  # y bias is less than 1e-6 pixel (in fine scale)


@pytest.mark.parametrize("psf_file,tolerance,rebin,npoints", [
    (GAUSS_PSF, 1e-7, 1, 10),  # centroiding still works ok with oversampled gaussians
    (GAUSS_PSF, 1e-7, 2, 10),  # on a small window
    (GAUSS_PSF, 1e-7, 2, 100)
])
def test_many_sources(psf_file, tolerance, rebin, npoints):
    """Verify that an image with a many sources can have their positions
    recovered to a desired accuracy.

    This function works by generating one image with floating point positioned
    sources and performing a CoG over sub-regions of the image.
    """
    psf_cog = get_psf_cog(psf_file) - np.array([0.5, 0.5])

    # make image and check cog
    src = make_source(n_sources=npoints)
    err_xl = []
    err_yl = []
    max_rad = 5.0  # arcsec
    pixsize = 3.75e-3  # arcsec
    cog_width = 128 // rebin  # pixels
    pos = max_rad * get_gr_distributed_sources(npoints)
    src.gauss_pos[:, :] = pos.copy()
    width_as = 15
    image_gen = ImageGenerator(array_width_pix=5000, source=src, psfs_file=psf_file,
                               pixsize=pixsize, gauss_width_pix=32, which_psf=0, norm_psf=True)
    image_gen.main()
    im = image_gen.get_rebinned_cropped(rebin_factor=rebin, cropped_width_as=width_as)
    for new_pos in pos:
        cog_offset = np.round((new_pos + width_as / 2) / pixsize - cog_width / 2).astype(np.int32) // rebin
        assert np.all(cog_offset > 0) and np.all((cog_offset + cog_width) < im.shape[0])
        cog_x, cog_y = get_cog_windowed(im, cog_offset[0], cog_offset[1], cog_width)
        # measure error in pixels:
        err_xl.append(cog_x * rebin - new_pos[0] / pixsize - psf_cog[0])
        err_yl.append(cog_y * rebin - new_pos[1] / pixsize - psf_cog[1])
    err_x = np.array(err_xl)
    err_y = np.array(err_yl)
    assert err_x.std() < tolerance     # x error std is less than 1e-6 pixel (in fine scale)
    assert err_y.std() < tolerance     # y error std is less than 1e-6 pixel (in fine scale)
    assert np.abs(err_x.mean()) < tolerance  # x bias is less than 1e-6 pixel (in fine scale)
    assert np.abs(err_y.mean()) < tolerance  # y bias is less than 1e-6 pixel (in fine scale)


@pytest.mark.parametrize("psf_file,rebin,npoints", [
    (GAUSS_PSF, 1, 10),
    (GAUSS_PSF, 2, 10),
    (E2E_PSF, 1, 10),
    (E2E_PSF, 2, 10),
    (E2E_PSF, 2, 100)
])
def test_batching_stars(psf_file, rebin, npoints):
    """verify that placing one star in the field at a time is the same as doing a batch
    of many stars all at once
    """
    # make image and check cog
    src = make_source(n_sources=npoints)
    max_rad = 5.0  # arcsec
    pixsize = 3.75e-3  # arcsec
    pos = max_rad * get_gr_distributed_sources(npoints)
    src.gauss_pos[:, :] = pos.copy()
    width_as = 15
    image_gen = ImageGenerator(array_width_pix=5000, source=src, psfs_file=psf_file,
                               pixsize=pixsize, gauss_width_pix=32, which_psf=0, norm_psf=True)
    image_gen.main()
    im_batched = image_gen.get_rebinned_cropped(rebin_factor=rebin, cropped_width_as=width_as)

    # make image and check cog
    src = make_source()
    im_stacked = np.zeros([4000 // rebin, 4000 // rebin])
    for new_pos in pos:
        src.gauss_pos[:] = new_pos.copy()
        image_gen = ImageGenerator(array_width_pix=5000, source=src, psfs_file=psf_file,
                                   pixsize=pixsize, gauss_width_pix=32, which_psf=0, norm_psf=True)
        image_gen.main()
        im_stacked += image_gen.get_rebinned_cropped(rebin_factor=rebin, cropped_width_as=width_as)

    assert np.allclose(im_batched, im_stacked)


# At this point in the tests, we can be convinced that:
#  - photometry is conserved (assuming the sampled PSFs capture all of the
#    flux in their windows),
#  - astrometry is conserved (up to a constant shift in the field which is
#    dictated by the conventions used)
#  - stacking many frames of single stars is the same as generating one field
#    with all of those stars at once.

def test_field_varyiability():
    pass


def test_polychromatic_photometry():
    pass


def test_polychromatic_astrometry():
    pass


def test_noise_statistics():
    pass



import numpy as np
import math
from scipy import interpolate
from astropy.convolution import convolve_fft
import poppy
from mavisim import generate_image
import astropy.io.fits as fits
from tqdm import tqdm


class AstromCalibSimGeneric():
    """Generic class for astrometric calibration simulation.
    See: AstromCalibSimAna for the analytical simulation
         AstromCalibSimE2E for the end-to-end simulation
    """

    def __init__(self, static_distort, *, pixel_size_as=0.00736,
                 pixel_size_mm=10e-3, dist_amp=1.0, mask_scale=1e3 / 582,
                 hole_position_std=0.0, dx=0.2, dy=0.2,
                 dx_meas=None, dy_meas=None, n_poly=6, pin_pitch=0.5,
                 num_pin_x=40):
        self._true_cam_samp = pixel_size_as                    # arcsec/pixel at output plane
        self._plate_scale = self._true_cam_samp / pixel_size_mm  # arcsec/mm at output plane
        self._static_distort = static_distort                  # static distortion data
        self._dist_amp = dist_amp                              # distortion amplification factor
        self._mask_scale = mask_scale                          # arcsec/mm at input plane
        self._pin_pitch = pin_pitch                            # mm at input plane
        self._num_pins = num_pin_x**2                          # 30x30 pinholes
        self._num_pin_x = num_pin_x                            # number of pins across the x-dimension (sqrt(num_pins_total))
        self._hole_position_std = hole_position_std
        self._n_poly = n_poly
        self._dx = dx
        self._dy = dy
        if dx_meas is not None:
            self._dx_meas = dx_meas
        else:
            self._dx_meas = dx
        if dy_meas is not None:
            self._dy_meas = dy_meas
        else:
            self._dy_meas = dy
        self._n_tot_poly = ((self._n_poly + 1) * (self._n_poly + 2)) // 2 - 1

        # Process input distortions into interpolated evaluatable function
        self._input_distortions_func = self._input_distortions()
        self._recovered_distortions_func = None
        self._p0_meas = None
        self._ppx_meas = None
        self._ppy_meas = None

    def input_dist(self, x, y):
        """Evaluate interpolated input distortions at arbitrary coordinates.

        `x` and `y` (in arcsec) can be anywhere in the science field, but must
        be array-like and the same size.

        Args:
            x : array-like float : field x-coordinates (arcsec)
            y : array-like float : field y-coordinates (arcsec)

        Returns
            out_x : array-like float : x-component of distortion at each coord
            out_y : array-like float : y-component of distortion at each coord
        """
        x = np.array(x).copy()
        y = np.array(y).copy()
        xx = x.flatten()
        yy = y.flatten()
        out_xx = xx * 0
        out_yy = xx * 0
        for i in range(xx.shape[0]):
            out_xx[i], out_yy[i] = self._input_distortions_func(xx[i], yy[i])[0]
        out_xx = out_xx.reshape(x.shape)
        out_yy = out_yy.reshape(x.shape)
        return out_xx, out_yy

    def _input_distortions_degmm(self):
        field_x = self._static_distort["Field_x(deg)"]
        field_y = self._static_distort["Field_y(deg)"]

        dist_x = self._static_distort["Predicted_x(mm)"] - self._static_distort["Real_x(mm)"]
        dist_y = self._static_distort["Predicted_y(mm)"] - self._static_distort["Real_y(mm)"]

        # Create an array of the field positions and the distortion at each pt (the difference)
        dist_all = np.empty([dist_x.shape[0], 4])
        dist_all[:, 0] = field_x
        dist_all[:, 1] = field_y
        dist_all[:, 2] = dist_x
        dist_all[:, 3] = dist_y

        grid_vals = np.unique(field_x)

        # Create grids to save the distortion (difference) in x and y at each point in the grid
        # This is necessary for the interpolation of the distortion (grid-wise interpolation)
        dist_x_grid = np.zeros([len(grid_vals), len(grid_vals)])
        dist_y_grid = np.zeros([len(grid_vals), len(grid_vals)])

        num = 0

        for x in grid_vals:
            sub_array = dist_all[np.where(dist_all[:, 0] == x), :][0]

            for row in np.arange(0, sub_array.shape[0]):

                dist_x_grid[num, row] = sub_array[row, 2]
                dist_y_grid[num, row] = sub_array[row, 3]

            num += 1

        dist_x_func_degmm = interpolate.RectBivariateSpline(grid_vals, grid_vals, dist_x_grid)
        dist_y_func_degmm = interpolate.RectBivariateSpline(grid_vals, grid_vals, dist_y_grid)

        return lambda x, y: np.c_[dist_x_func_degmm(x, y), dist_y_func_degmm(x, y)] * self._dist_amp

    def _input_distortions(self):
        """Distortion function generator wrapper for arcsec to arcsec.
        Call this function to generate
        """
        out_func = self._input_distortions_degmm()
        return lambda x, y: out_func(x / 3600, y / 3600) * self._plate_scale

    def recovered_dist(self, x, y):
        """Evaluate recovered/estimated input distortions at arbitrary coordinates.
        This is the estimated distortion via the differential calibration method
        based on the input static distortion.

        `x` and `y` (in arcsec) can be anywhere in the science field, but must
        be array-like and the same size.

        Args:
            x : array-like float : field x-coordinates (arcsec)
            y : array-like float : field y-coordinates (arcsec)

        Returns
            out_x : array-like float : x-component of distortion at each coord
            out_y : array-like float : y-component of distortion at each coord
        """
        if self._recovered_distortions_func is None:
            raise ValueError("Generic astrometric simulator does not implement distortion recovery method")
        x = np.array(x).copy()
        y = np.array(y).copy()
        xx = x.flatten()
        yy = y.flatten()
        out_xx = xx * 0
        out_yy = xx * 0
        for i in range(xx.shape[0]):
            out_xx[i], out_yy[i] = self._recovered_distortions_func(xx[i], yy[i])[0]
        out_xx = out_xx.reshape(x.shape)
        out_yy = out_yy.reshape(x.shape)
        return out_xx, out_yy

    def residual_dist(self, x, y):
        """Evaluate residual distortions at arbitrary coordinates.

        `x` and `y` (in arcsec) can be anywhere in the science field, but must
        be array-like and the same size.

        Args:
            x : array-like float : field x-coordinates (arcsec)
            y : array-like float : field y-coordinates (arcsec)

        Returns
            out_x : array-like float : x-component of distortion at each coord
            out_y : array-like float : y-component of distortion at each coord
        """
        if self._recovered_distortions_func is None:
            raise ValueError("Generic astrometric simulator does not implement distortion recovery method")

        x = np.array(x).copy()
        y = np.array(y).copy()
        xx = x.flatten()
        yy = y.flatten()
        out_xx = xx * 0
        out_yy = xx * 0
        for i in range(xx.shape[0]):
            out_xx[i], out_yy[i] = self._input_distortions_func(xx[i], yy[i])[0] - \
                self._recovered_distortions_func(xx[i], yy[i])[0]
        out_xx = out_xx.reshape(x.shape)
        out_yy = out_yy.reshape(x.shape)
        return out_xx, out_yy

    @staticmethod
    def _hbvpoly(p, a, n_poly):
        """ Evaluate the homogenous bi-variate polynomial defined by
        coefficients in a at position p.

        Arguments:
            p: np.ndarray : position to evaluate polynomial at, (M,2)
            a: np.ndarray : coefficients defining polynomial, (((N+1)(N+2))//2-1,)
            N: int: maximum homogenous polynomial order to go to.

        Returns:
            out: np.ndarray : evaluated polynomial, scalar or (M,1)
        """
        if len(p.shape) != 2:
            raise ValueError("p must be 2D, i.e., p.shape=(M,2)")
        out = np.zeros_like(p[:, 0])
        counter = 0
        for n in range(1, n_poly + 1):  # skip tip-tilt
            for j in range(n + 1):
                out[:] += a[counter] * p[:, 0]**j * p[:, 1]**(n - j) / math.factorial(n)
                counter += 1
        return out

    @staticmethod
    def _hbvpoly_grad(p, n_poly):
        """ Evaluate the gradient of the homogenous bi-variate polynomial
        defined by coefficients in a at position p.

        Arguments:
            p: np.ndarray : position to evaluate polynomial gradient at, (2,) or (M,2)
            n_poly: int: maximum homogenous polynomial order to go to.

        Returns:
            out: np.ndarray : evaluated polynomial gradient,

        """
        dx = np.zeros([p.shape[0], ((n_poly + 1) * (n_poly + 2)) // 2 - 1])
        dy = np.zeros([p.shape[0], ((n_poly + 1) * (n_poly + 2)) // 2 - 1])
        counter = -1
        for n in range(1, n_poly + 1):  # skip tip-tilt
            for j in range(n + 1):
                counter += 1
                if j == 0:
                    continue
                dx[:, counter] += j * p[:, 0]**(j - 1) * p[:, 1]**(n - j) / math.factorial(n)

        counter = -1
        for n in range(1, n_poly + 1):  # skip tip-tilt
            for j in range(n + 1):
                counter += 1
                if j == n:
                    continue
                dy[:, counter] += (n - j) * p[:, 0]**j * p[:, 1]**(n - j - 1) / math.factorial(n)
        return dx, dy

    @staticmethod
    def _make_pinhole_grid(xshift=0., yshift=0., sigma=0., grid="square", incl_dist=True,
                           pins_per_side=30, mask_scale=1e3 / 582, pin_pitch=0.571,
                           plate_scale=7.36e-3 / 10e-3, dist_func_degmm=None, seed=1234):
        """
        Generate arrays of x-y pinhole positions in pixels and arcsec.
        Optionally pass a global shift in x/y, in mm, to shift pinhole grid
        relative to distortion field. Can also provide an uncertainty in the hole
        positions (also in mm), which is treated as Gaussian. Distortions are
        included by default, but can be turned off to get "nominal" pinhole grid.


        Args:
            xshift (float, optional): Shift amount in x-axis (arcsec). Defaults to 0..
            yshift (float, optional): Shift amount in y-axis (arcsec). Defaults to 0..
            sigma (float, optional): Standard deviation on pinhole positions. Defaults to 0..
            incl_dist (bool, optional): Flag to include distortions in returned coordinates. Defaults to True.
            pins_per_side (int, optional): Number of pins per side of the grid. Defaults to 30.
            mask_scale (float, optional): arcsec/mm at the mask. Defaults to (1e3/582).
            pin_pitch (float, optional): spacing between the pinholes (in mm). Defaults to 0.582.
            plate_scale (float, optional): arcsec/mm at the sensor. Defaults to 7.36e-3/10e-3.
            dist_func_degmm (function, optional): Function to obtain distortion.
                Takes argument in degrees in field, returns distortion in mm at sensor. Defaults to None.

        Returns:
            numpy.ndarray: 2d array of x-y pinhole positions in arcsec.
        """
        if grid == "square":
            # individual hole positions in mm (relative to centre)
            x_mm, y_mm = np.meshgrid(
                (np.arange(pins_per_side) - (pins_per_side - 1) / 2.) * pin_pitch,
                (np.arange(pins_per_side) - (pins_per_side - 1) / 2.) * pin_pitch
            )
        elif grid == "hex":
            n_init = pins_per_side
            spacing = pin_pitch
            x_mm, y_mm = np.meshgrid(np.linspace(-n_init / 2, n_init / 2, n_init + 1),
                                     np.linspace(-n_init / 2, n_init / 2, n_init + 1))
            y_mm[:, ::2] += 1 / 2
            x_mm *= (np.sqrt(3) / 2)
            y_mm -= 0.5
            x_mm *= spacing
            y_mm *= spacing
        else:
            raise ValueError("grid type: " + grid + " is unsupported.")
        mask = (np.abs(x_mm) <= 15.0/mask_scale) * (np.abs(y_mm) <= 15.0 / mask_scale)
        x_mm = x_mm[mask]
        y_mm = y_mm[mask]
        x_mm = x_mm.flatten()
        y_mm = y_mm.flatten()

        pin_loc_mm_in = np.c_[x_mm, y_mm]

        # apply any shift terms
        pin_loc_mm_in[:, 0] += xshift
        pin_loc_mm_in[:, 1] += yshift

        valid = (pin_loc_mm_in**2).sum(axis=1)**0.5 <= (14.5 / mask_scale)

        # include uncertainty in hole position?
        np.random.seed(seed)
        pin_loc_mm_in += sigma * np.random.randn(*pin_loc_mm_in.shape)

        # convert from mm to deg. Use full grid to include individual point uncertainties
        pin_loc_as = pin_loc_mm_in * mask_scale

        if incl_dist:
            if dist_func_degmm is None:
                raise ValueError("Must provide distortion functions to include distortion")
            # get distortion terms at each location
            for ii, loc_deg in enumerate(pin_loc_as):
                pin_loc_as[ii, :] += plate_scale * dist_func_degmm(loc_deg[0] / 3600, loc_deg[1] / 3600).flatten()

        return pin_loc_as, valid

    def _fit_poly(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        n_pos = self._valid.sum()

        d_mat = np.zeros([4 * n_pos, 2 * self._n_tot_poly])
        grad_tmp = self._hbvpoly_grad(self._p0_nom[self._valid], self._n_poly)

        d_mat[0::4, :self._n_tot_poly] = grad_tmp[0]
        d_mat[1::4, self._n_tot_poly:] = grad_tmp[0]
        d_mat[2::4, :self._n_tot_poly] = grad_tmp[1]
        d_mat[3::4, self._n_tot_poly:] = grad_tmp[1]

        d_inv = np.linalg.solve(d_mat.T @ d_mat, d_mat.T)

        dx_arcsec = self._mask_scale * self._dx_meas
        dy_arcsec = self._mask_scale * self._dy_meas

        # compenent-wise gradients:
        if self._p0_meas is None:
            raise RuntimeError("measurements must be made before fitting polynomial")
        dpdx = (self._ppx_meas - self._p0_meas) - np.r_[dx_arcsec, 0]
        dpdx /= dx_arcsec
        dpdy = (self._ppy_meas - self._p0_meas) - np.r_[0, dy_arcsec]
        dpdy /= dy_arcsec

        # estimated gradients:
        z_hat = np.c_[dpdx[self._valid], dpdy[self._valid]].flatten()
        # estimated polynomial coefficients:
        return d_inv @ z_hat

    def _set_coords_square(self):
        kwargs_grid = {
            "pins_per_side": self._num_pin_x,
            "mask_scale": self._mask_scale,
            "pin_pitch": self._pin_pitch,
            "plate_scale": self._plate_scale,
            "dist_func_degmm": self._input_distortions_degmm(),
            "grid": "square"
        }
        self._p0_nom, v = self._make_pinhole_grid(incl_dist=False, **kwargs_grid)
        self._ppx_nom, v_ = self._make_pinhole_grid(xshift=self._dx, incl_dist=False, **kwargs_grid)
        v = v * v_
        self._ppy_nom, v_ = self._make_pinhole_grid(yshift=self._dy, incl_dist=False, **kwargs_grid)
        v = v * v_
        self._valid = v
        self._p0_true, _ = self._make_pinhole_grid(sigma=self._hole_position_std, **kwargs_grid)
        self._ppx_true, _ = self._make_pinhole_grid(xshift=self._dx, sigma=self._hole_position_std, **kwargs_grid)
        self._ppy_true, _ = self._make_pinhole_grid(yshift=self._dy, sigma=self._hole_position_std, **kwargs_grid)

    def _set_coords_hex(self):
        kwargs_grid = {
            "pins_per_side": self._num_pin_x,
            "mask_scale": self._mask_scale,
            "pin_pitch": self._pin_pitch,
            "plate_scale": self._plate_scale,
            "dist_func_degmm": self._input_distortions_degmm(),
            "grid": "hex"
        }
        self._p0_nom, v = self._make_pinhole_grid(incl_dist=False, **kwargs_grid)
        self._ppx_nom, v_ = self._make_pinhole_grid(xshift=self._dx, incl_dist=False, **kwargs_grid)
        v = v * v_
        self._ppy_nom, v_ = self._make_pinhole_grid(yshift=self._dy, incl_dist=False, **kwargs_grid)
        v = v * v_
        self._valid = v
        self._p0_true, _ = self._make_pinhole_grid(sigma=self._hole_position_std, **kwargs_grid)
        self._ppx_true, _ = self._make_pinhole_grid(xshift=self._dx, sigma=self._hole_position_std, **kwargs_grid)
        self._ppy_true, _ = self._make_pinhole_grid(yshift=self._dy, sigma=self._hole_position_std, **kwargs_grid)

###############################################################################
###############################################################################
###############################################################################


class AstromCalibSimAna(AstromCalibSimGeneric):
    """
    """

    def __init__(self, *args, centroid_noise_std=0.0, **kwargs):
        super().__init__(*args, **kwargs)
        self._centroid_noise_std = centroid_noise_std
        # Perform calibration process, and process recovered distortions into evaluatable function
        self._recovered_distortions_func = self._recovered_distortions_ana()

    def _do_measurements(self):
        rand_shape = self._p0_nom.shape
        self._p0_meas = self._p0_true + np.random.randn(*rand_shape) * self._centroid_noise_std
        self._ppx_meas = self._ppx_true + np.random.randn(*rand_shape) * self._centroid_noise_std
        self._ppy_meas = self._ppy_true + np.random.randn(*rand_shape) * self._centroid_noise_std

    def _recovered_distortions_ana(self):
        self._set_coords_square()
        self._do_measurements()
        u_hat = self._fit_poly()
        return lambda x, y: np.c_[self._hbvpoly(np.c_[x, y], u_hat[:self._n_tot_poly], self._n_poly),
                                  self._hbvpoly(np.c_[x, y], u_hat[self._n_tot_poly:], self._n_poly)]

###############################################################################
###############################################################################
###############################################################################


class AstromCalibSimE2E(AstromCalibSimGeneric):
    """
    """

    def __init__(self, *args, pin_size=10e-3, pinhole_os=4, pixel_os=2, wavelength=550e-9, pinhole_support_width=128,
                 noise_fun=None, centroid_win_rad=0.2, centroid_threshold=0.0, **kwargs):
        super().__init__(*args, **kwargs)
        self._pin_size = pin_size           # pinhole physical diameter in mm
        self._pinhole_os = pinhole_os       # oversampling factor
        self._pixel_os = pixel_os           # pixel oversampling factor
        self._wavelength = wavelength       # wavelength in metres
        self._fine_cam_samp = self._true_cam_samp / pixel_os
        self._noise_fun = noise_fun
        self._centroid_win_rad = centroid_win_rad
        self._centroid_threshold = centroid_threshold
        self._pinhole_support_width = pinhole_support_width

        # Perform calibration process, and process recovered distortions into evaluatable function
        self._recovered_distortions_func = self._recovered_distortions_e2e()

    def _recovered_distortions_e2e(self):
        self._set_coords_hex()
        self._do_measurements()
        u_hat = self._fit_poly()
        return lambda x, y: np.c_[self._hbvpoly(np.c_[x, y], u_hat[:self._n_tot_poly], self._n_poly),
                                  self._hbvpoly(np.c_[x, y], u_hat[self._n_tot_poly:], self._n_poly)]

    def _centroids(self, pos, im, origin):
        """_summary_

        Args:
            pos (_type_): _description_
            im (_type_): _description_
            win_rad (float, optional): _description_. Defaults to 0.2.

        Returns:
            _type_: _description_
        """
        cog_meas = []
        win_rad = self._centroid_win_rad
        # iterate over pinholes:
        for pin_x, pin_y in tqdm(pos, desc="centroids", leave=False):
            # integer-valued indices for desired pinhole-window
            win_idx = np.mgrid[int((pin_y - win_rad - origin[1]) / self._true_cam_samp):
                               int((pin_y + win_rad - origin[1]) / self._true_cam_samp) + 1,
                               int((pin_x - win_rad - origin[0]) / self._true_cam_samp):
                               int((pin_x + win_rad - origin[0]) / self._true_cam_samp) + 1]
            win_idx = np.clip(win_idx, 0, im.shape[0] - 1)
            # corresponding pixel coordinates within that window
            win_as = [((w+0.5) * self._true_cam_samp + origin[0]).flatten() for w in win_idx]
            # pixel intensities of window, flattened for centroiding
            window = im[win_idx[0], win_idx[1]].flatten()
            # centroid calculation
            cog_meas.append([(window @ p) / window.sum() for p in win_as][::-1])
        return np.r_[cog_meas]

    def _do_measurements(self):
        # Do this by making an image and measuring the positions:
        pin_image = self._pinhole_image()
        pin_image = pin_image / (self._pixel_os**2)
        pin_image_filename = "_pin_image.fits"
        fits.writeto(pin_image_filename, np.array([[]]), overwrite=True)
        fits.append(pin_image_filename, pin_image,
                    fits.Header({"YPOS": 0.0, "XPOS": 0.0, "LAMBDA": self._wavelength}))

        class SourceHack:
            def __init__(self, coords):
                self.flux = 1.0 * np.ones(len(coords))
                self.gauss_pos = coords.copy()
                self.cov_mat = None

        source_p0 = SourceHack(self._p0_true)
        source_ppx = SourceHack(self._ppx_true)
        source_ppy = SourceHack(self._ppy_true)

        image_gen_p0 = generate_image.ImageGenerator(6400 * self._pixel_os,
                                                     source_p0, pin_image_filename,
                                                     self._fine_cam_samp, which_psf=0, norm_psf=False)
        image_gen_ppx = generate_image.ImageGenerator(6400 * self._pixel_os,
                                                      source_ppx, pin_image_filename,
                                                      self._fine_cam_samp, which_psf=0, norm_psf=False)
        image_gen_ppy = generate_image.ImageGenerator(6400 * self._pixel_os,
                                                      source_ppy, pin_image_filename,
                                                      self._fine_cam_samp, which_psf=0, norm_psf=False)
        image_gen_p0.main()
        image_gen_ppx.main()
        image_gen_ppy.main()

        self._im0 = image_gen_p0.get_rebinned_cropped(self._pixel_os, self._true_cam_samp * 4000)
        self._impx = image_gen_ppx.get_rebinned_cropped(self._pixel_os, self._true_cam_samp * 4000)
        self._impy = image_gen_ppy.get_rebinned_cropped(self._pixel_os, self._true_cam_samp * 4000)

        if self._noise_fun is not None:
            self._im0 = self._noise_fun(self._im0)
            self._impx = self._noise_fun(self._impx)
            self._impy = self._noise_fun(self._impy)

        self._im0 -= self._centroid_threshold
        self._impx -= self._centroid_threshold
        self._impy -= self._centroid_threshold
        self._im0[self._im0 < 0] = 0
        self._impx[self._impx < 0] = 0
        self._impy[self._impy < 0] = 0

        self._p0_meas = np.zeros(self._p0_nom.shape)
        self._ppx_meas = np.zeros(self._ppx_nom.shape)
        self._ppy_meas = np.zeros(self._ppy_nom.shape)

        # TODO: this origin thing is ugly and maybe buggy, can be cleaned up with a convention change
        origin = np.array([-0.5, -0.5]) * (self._true_cam_samp * 4000)
        self._p0_meas[self._valid] = self._centroids(pos=self._p0_nom[self._valid], im=self._im0, origin=origin)
        self._ppx_meas[self._valid] = self._centroids(pos=self._ppx_nom[self._valid], im=self._impx, origin=origin)
        self._ppy_meas[self._valid] = self._centroids(pos=self._ppy_nom[self._valid], im=self._impy, origin=origin)

    @staticmethod
    def _pinhole(size, x, y, radius):
        """generate the sampled pinhole function

        Args:
            size (int): size of the output array (in pixels)
            x (float): x position of the pinhole (in pixels)
            y (float): y position of the pinhole (in pixels)
            radius (float): radius of the pinhole (in pixels)

        Returns:
            numpy.ndarray: 2d array of the pinhole function
        """
        # generate coordinates:
        pos = np.mgrid[:size, :size]
        # signed distance to pinhole edge:
        dr = np.sqrt((pos[1] - x)**2 + (pos[0] - y)**2) - radius
        # generate the pinhole function:
        weight = 0.5 - np.clip(dr, -0.5, 0.5)
        return weight

    def _pinhole_image(self):
        """generate the pinhole image

        Args:
            output_size (int): size of the output array (in pixels)

        Returns:
            numpy.ndarray: 2d array of the pinhole image
        """
        # Pinhole Representation, fourier transform the pinhole directly
        pin_ap = (self._pin_size * self._mask_scale) * self._pinhole_os / self._fine_cam_samp
        output_size = self._pinhole_support_width
        # 2d point grid
        newsize = output_size * self._pinhole_os
        centre = newsize / 2

        # make a pinhole
        pin = self._pinhole(newsize, centre, centre, pin_ap)

        # generate the PSF with poppy
        osys = poppy.OpticalSystem()
        osys.add_pupil(poppy.CircularAperture(radius=4.))    # pupil radius in meters
        osys.add_detector(pixelscale=self._fine_cam_samp, fov_pixels=output_size, oversample=self._pinhole_os)
        psf = osys.calc_psf(self._wavelength)
        tel_psf = psf[0].data / psf[0].data.sum()

        # smoosh them together
        pin_image = convolve_fft(pin, tel_psf, allow_huge=True)

        # bin back down
        rb = np.copy(pin_image).reshape(
            output_size, self._pinhole_os, output_size, self._pinhole_os
        ).sum(axis=(-1, 1)) / self._pinhole_os**2

        return rb



from astropy.io import ascii
from astropy import units as u
import numpy as np

####
# Input Catalogue info. BYO Catalogue here:
input_cat = ascii.read("tests/test_cat")

####
# Gaussian Point Source Info. E.g., to investigate sensitivity to vibrations
vib_term = 0.48  # 0.48 pixels, FWHM = 8.225 mas extra vibration, accounting for the 10mas min FWHM
cd_term = 0.4  # 0.4 pixels, FWHM = 7.05 pixels to account for charge diffusion, base jitter in the case of a static jitter kernal
gauss_offset = 0.0  # pixels, offset to make sure the rebinning maintains a centred PSF

####
# Static Distortion. E.g., to investigate sensitivity to VLT+MAVIS distortions
plate_scale = 1.29  # arcsec/mm, to convert location in mm to arcseconds
dynamic_amp = 1  # amplification factor to increase the distortion

####
# Sky Background
surf_bright = 21.61   # mag/square arcsecond

####
# MAVIS specific. Unlikely to be modified by the user:

# CCD info
ccd_size = 4000  # Change this to 4096 for FFT speed
ccd_sampling = 0.0075  # Pixel/arcsec, should be the same as the sampled FV PSF prior to convolution

# noise related parameters
psf_wavelength = 550   # nm (monochromatic version)
filt_width = 88        # nm, width of V band
collecting_area = (np.pi * 4.0**2 * (1 - 0.16**2)) * u.m**2

# FoV for trimming input catalogue
MAVIS_fov = 30  # arcseconds
buffer = 10.005  # buffer in arcseconds to add to the MAVIS fov to account for stars outside the science fov

# Detector Characteristics
slow_rdnoise = 3  # electrons, best
fast_rdnoise = 5  # electrons, worst
bit_depth = 16  # standard
sat_point = 0  # could specificy this instead
gain = 1.0  # taken fron Francois ADU/e
noise_mode = "slow"
AOMthruput = 0.63  # for 550 nm --> generalise this!
VLTthruput = 0.75  # approx flat across all wavelengths
QE = 0.89 		   # for 550nm  --> generalise this



from .util import make_static_dist_map, input_coo, add_all_noise
from .generate_image import ImageGenerator
from .source import Source
from .astromsim import AstromCalibSimAna
from .astromsim import AstromCalibSimAna as AstromCalibSim
from .astromsim import AstromCalibSimE2E

__all__ = ["make_static_dist_map",
           "input_coo",
           "add_all_noise",
           "ImageGenerator",
           "Source",
           "AstromCalibSim",
           "AstromCalibSimAna",
           "AstromCalibSimE2E"]



# Module to calculate noise using the Pyxel tool

import pyxel
from pyxel.util import fit_into_array
import numpy as np
import numpy.typing as npt
from pyxel.detectors import CCD, APD


def calc_noise_pyxel(
    input_photon: npt.NDArray,
    yaml_file: str,
) -> np.ndarray | None:
    """ Takes photon array and yaml parameter file name and processes it to add noise

    Parameters
    ----------
    input_photon: ndarray (pass photon array)
    yaml_file: yaml file containing detector properties

    Returns
    -------
    output_image: ndarray

    """
    # Add pyxel noise calculation stuff here
    # Make sure that the photon only has positive elements
    input_photon[input_photon < 0.0] = 0.0
    row, col = np.shape(input_photon)

    # Get Pyxel configs from the yaml file

    config = pyxel.load(yaml_file=yaml_file)

    if config.exposure is None:
        return None

    # Set pyxel stuff from the config file
    exposure = config.exposure  # class Single
    pipeline = config.pipeline  # class DetectionPipeline

    # Currently this only supports ccd and apd detector,
    # but it's easy to add support for other detectors

    detector: APD | CCD
    # set detector to ccd or apd
    if (config.ccd_detector is not None):
        detector = config.ccd_detector
    elif (config.apd_detector is not None):
        detector = config.apd_detector
    else:
        raise ValueError("Failed to set detector. Must be either apd or ccd.")

    # Get detector size
    detector_row = detector.geometry.row
    detector_col = detector.geometry.col

    # Pad zeros to array to match detector size
    det_shape = (detector_row, detector_col)
    padded_pyxel_photon_array = fit_into_array(array=input_photon,
                                               output_shape=det_shape, align='center')

    photon_array = padded_pyxel_photon_array
    try:
        detector.photon.array += photon_array
    except ValueError as ex:
        raise ValueError("Shapes of arrays do not match") from ex

    # run pyxel in exposure mode
    if exposure:
        pyxel_result = pyxel.exposure_mode(
            exposure=exposure, detector=detector, pipeline=pipeline)
    else:
        raise NotImplementedError

    # Get the image from Pyxel
    if config.exposure:
        pyxel_readout = config.exposure.readout.times[0]
        output_image_padded = np.array(pyxel_result.image.sel(readout_time=pyxel_readout))

    # Remove padding
    output_image = fit_into_array(
        array=output_image_padded,
        output_shape=(row, col), align='center'
    )

    return np.array(output_image)
